/***************************************************************************
    copyright            : (C) 1999 by Frederik Bilhaut
    email                : bilhaut_f@mail.cpod.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "pluginmanager.h"
#include <qmessagebox.h>
#include <qdir.h>
#include <qfileinfo.h>
#include <kapp.h>

PluginManager *PluginManager::instance = 0;

PluginManager::PluginManager(void)
	{
	if(instance == 0)
		instance = this;
	else
		warning("Multiple instances of PluginManager not allowed!");
	for(int i=0; i<MAX_PLUGINS; i++)
		plugins[i] = 0;
	}
	
PluginManager::~PluginManager(void)
	{
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i])
			unregisterPlugin(i);
	}
	
PluginManager *PluginManager::getInstance(void)
	{
	return instance;
	}
	
int PluginManager::registerPlugin(QString path)
	{
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i] == 0)
			{
			Plugin *plugin = new Plugin;
			if(!plugin->load(path))
				{
				errorString = plugin->getErrorString();
				return plugin->getErrorCode();
				}
			if(getPluginByName(plugin->getPluginName()) >= 0)
				{
				plugin->unload();
				errorString = "Duplicate plugin";
				return -10;
				}
			plugins[i] = plugin;
			warning("Register plugin: %s -> id %d", (const char *)path, i);
			return i;
			}
	errorString = "Maximum number of plugins reached (how did you do that ?)";
	return -11;
	}

int PluginManager::unregisterPlugin(int id)
	{
	if(id >= MAX_PLUGINS)
		return -1;
	if(!plugins[id])
		return -2;
	if(!plugins[id]->unload())
		return -3;
	delete plugins[id];
	plugins[id] = 0;
  warning("Unregister plugin: %d", id);
	return 0;
	}
	
ISqlServer *PluginManager::createInstance(int id)
	{
	if(id >= MAX_PLUGINS)
		return 0;
	Plugin *plugin = plugins[id];
	if(!plugin)
		return 0;
	IObject *instance = plugin->createInstance();
	if(!instance)
		return 0;
	ISqlServer *iSql = (ISqlServer *)(instance->queryInterface(UIID_SQLSERVER));
	if(!iSql)
		return 0;
	return iSql;
	}
		
const char *PluginManager::getPluginName(int id)
	{
	if(id >= MAX_PLUGINS)
		return 0;
	Plugin *plugin = plugins[id];
	if(!plugin)
		return 0;
	return plugin->getPluginName();
	}
	
const char *PluginManager::getPluginInfo(int id)
	{
	if(id >= MAX_PLUGINS)
		return 0;
	Plugin *plugin = plugins[id];
	if(!plugin)
		return 0;
	return plugin->getPluginInfo();
	}

const char *PluginManager::getPluginPath(int id)
	{
	if(id >= MAX_PLUGINS)
		return 0;
	Plugin *plugin = plugins[id];
	if(!plugin)
		return 0;
	return plugin->getPluginPath();
	}
	
int PluginManager::getPluginByName(QString name)
	{
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i] && name == plugins[i]->getPluginName())
			return i;
	return -1;
	}

int PluginManager::getPluginByPath(QString path)
	{
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i] && QFileInfo(path).absFilePath() == QFileInfo(plugins[i]->getPluginPath()).absFilePath())
			return i;
	return -1;
	}

QStrList *PluginManager::getPluginNames(void)
	{
	QStrList *res = new QStrList;
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i])
			res->append(plugins[i]->getPluginName());
	return res;
	}


void PluginManager::savePref(KConfig *config)
	{
	// Save all modules
	config->setGroup("Modules");
	int count = 0;
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i])
			count++;
	config->writeEntry("Count", count);
	int j = 0;
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i])
			{
			QString nr;
			nr.sprintf("%d", j);
			config->writeEntry(nr, plugins[i]->getPluginPath());
			j++;
			}	
	}



void PluginManager::readPref(KConfig *config)
	{
	config->setGroup("Modules");
	int count = config->readNumEntry("Count", 0);
	for(int i=0; i<count; i++)
		{
		QString nr;
		nr.sprintf("%d", i);
		QString pluginPath = config->readEntry(nr);
		int id = registerPlugin(pluginPath);
		if(id<0)
    	{
    	QString msg(512);
    	msg.sprintf("Error %d while loading plugin:\n%s", id, getErrorString());
    	QMessageBox::critical(0, "Error", msg);
    	return;
    	}	
		}
	}

int PluginManager::count(void)
	{
	int res = 0;
	for(int i=0; i<MAX_PLUGINS; i++)
		if(plugins[i])
			res++;
	return res;
	}



bool PluginManager::autodetectPlugins()
	{
	QDir dir(KApplication::kde_bindir());
	if(!dir.cd("../lib/kmysql/plugins"))
		{
		warning("Plugin auto-detection: directory $KDEDIR/lib/kmysql/plugins not found!");
		return false;
		}
	warning("Plugin auto-detection: looking in %s", (const char *)dir.absPath());
	dir.setFilter(QDir::Files|QDir::Readable|QDir::Executable);
	dir.setNameFilter("*.so");
	const QStrList *foundPlugins = dir.entryList();
	if(foundPlugins == 0)
		{
		warning("Plugin auto-detection: can't list directory %s", (const char *)dir.absPath());
		return false;
		}
	warning("Plugin auto-detection: %d plugin(s) found", foundPlugins->count());	
	QString msg = i18n("The following plugins have been found and installed:\n");
	bool installed = false;
	for(QStrListIterator it(*foundPlugins); it; ++it)
		{
		QString path = dir.absPath()+"/"+it.current();
		warning("Found plugin: %s", (const char *)path);	
		if(getPluginByPath(path) >= 0)
			{
			warning("Plugin already intalled: %s", (const char *)path);
			continue;
			}
		// We register. It will be aborted if the plugin already exists.
		int res = registerPlugin(path);
		if(res == -10)
			warning("Plugin not installed (Duplicate plugin)");
		else if(res < 0)
			warning("Error %d while installing plugin %s", res, (const char *)path);
		else
			{
			installed = true;
			msg += "\n" + path;
			}			
		}
	if(installed){
	//	QMessageBox::information(KApplication::getKApplication()->topWidget(), i18n("Plugin auto-detection"), msg);
	}
	return true;
	}





