/***************************************************************************
    copyright            : (C) 1999 by Frederik Bilhaut
    email                : bilhaut_f@mail.cpod.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tableeditorview.h"
#include "mainwindow.h"
#include "utils.h"

TableEditorView::TableEditorView(MainWindow *parent)
	: View(parent)
	{
	setFrameStyle(NoFrame);
	initWidgets();
	initToolBars();
	initMenus();
	initConnects();
	}
	
void TableEditorView::initWidgets(void)
	{
	QHBoxLayout *layout = new QHBoxLayout(this);
	tab = new TabView(this);
	layout->addWidget(tab);
	tab->setEditable(true);
	}
	
void TableEditorView::initToolBars(void)
	{
	KToolBar *tbTe = new KToolBar(mainWindow());
	KIconLoader *iconLoader = KApplication::getKApplication()->getIconLoader();	
	QPixmap pixTe1 = iconLoader->loadIcon(QString("tick.xpm"));
  tbTe->insertButton(pixTe1, 1, SIGNAL(clicked()), mainWindow()->mainView(), SLOT(slotTableEditDone()), true, "Done");
	tbId = mainWindow()->addToolBar(tbTe);
	}
	
void TableEditorView::initMenus(void)
	{
	KIconLoader *iconLoader = KApplication::getKApplication()->getIconLoader();
	QPopupMenu *table = new QPopupMenu();
	table->insertItem(iconLoader->loadIcon("tick.xpm"), i18n("Done"), mainWindow()->mainView(), SLOT(slotTableEditDone()), CTRL+Key_S);
 	menu = new KMenuBar(mainWindow());
  menu->insertItem(i18n("&Table"), table);
	}
  	
void TableEditorView::initConnects(void)
	{
	connect(tab, SIGNAL(cellChanged(int, int, const char *, const char *)), SLOT(slotCellChanged(int, int, const char *, const char *)));
	}

TableEditorView::~TableEditorView(void)
	{
	}

	
void TableEditorView::setEnabled(bool enabled)
	{
	if(enabled)
		{
		show();
		menu->show();
		mainWindow()->setMenu(menu);
		mainWindow()->enableToolBar(KToolBar::Show, tbId);
		mainWindow()->setView(this);
		}
	else
		{
		hide();
		menu->hide();
		mainWindow()->enableToolBar(KToolBar::Hide, tbId);
		}
	}

void TableEditorView::edit(ISqlServer *_sql, QString _tableName)
	{
	sql = _sql;
	tableName = _tableName;
  QString query = "SELECT * FROM " + tableName;
	QueryResult res = sql->query(query);
  if(res.type == QueryResult::Error)
    {
    QMessageBox::critical(this, i18n("Error"), res.message);
    return;
    }
  tab->clear();
  tab->setAutoUpdate(false);
  int nbcols = res.numColumns;
  tab->setNumCols(nbcols);
  tab->setColumns(res.fields);
  QStrList rows = res.result;
  for(unsigned int i=0; i<rows.count(); i++)
  	tab->appendItem(rows.at(i));
  tab->fitAll();
  tab->update();
  tab->setAutoUpdate(true);
  tab->repaint();
  mainWindow()->statusMessage(i18n("Editing table ")+tableName);
	}
	
void TableEditorView::slotCellChanged(int row, int col, const char *oldValue, const char *newValue)
	{
	QStrList keys;
	if(!getKeys(keys))
		{
		QMessageBox::critical(this, "Error", "Cannot get keys!");
		return;
		}
	if(keys.isEmpty())
		{
		tab->changeItemPart(oldValue, row, col);
		QMessageBox::warning(this, "Error", "You can't edit this table since it has no primary key.");
		return;
		}
	QString update = "UPDATE "+tableName+" SET "+tab->getHeadText(col)+" = '"+Utils::insertQuotes(newValue)+"' WHERE ";
	QString select = "SELECT * FROM "+tableName+" WHERE ";
	bool somethingBefore = false;
	for(int i=0; i<tab->numCols(); i++)
		{
		if(keys.find(tab->getHeadText(i)) == -1)
			continue;
		if(somethingBefore)
			{
			update += " AND ";
			select += " AND ";
			}
		select += tab->getHeadText(i)+" = ";
		if(i == col)
			select += "'"+Utils::insertQuotes(newValue)+"'";
		else
			select +="'"+Utils::insertQuotes(QString(tab->text(row, i)).stripWhiteSpace())+"'";
		update += tab->getHeadText(i)+" = ";
		if(i == col)
			update += "'"+Utils::insertQuotes(oldValue)+"'";
		else
			update +="'"+Utils::insertQuotes(QString(tab->text(row, i)).stripWhiteSpace())+"'";
		somethingBefore = true;
		}
	warning("%s", (const char *)update);
	QueryResult res = sql->query(update);
	if(res.type == QueryResult::Error)
    {
    QMessageBox::critical(this, "Error", res.message);
    tab->changeItemPart(oldValue, row, col);
    return;
    }
 	warning("%s", (const char *)select);
	res = sql->query(select);
	if(res.type == QueryResult::Error)
    {
    QMessageBox::critical(this, "Error", res.message);
    return;
    }
	tab->changeItem(res.result.at(0), row);
	}
	

bool TableEditorView::getKeys(QStrList &keys)
	{
	QString query = "SHOW KEYS FROM "+tableName;
	QueryResult res = sql->query(query);
	if(res.type == QueryResult::Error)
    {
    QMessageBox::critical(this, i18n("Error"), res.message);
    return false;
    }
	keys.clear();
	for(unsigned int i=0; i<res.result.count(); i++)
		{
		QString keyName = Utils::subStringAt(res.result.at(i), '\t', 2);
		QString columnName = Utils::subStringAt(res.result.at(i), '\t', 4);
		if(keyName == "PRIMARY")
			keys.append(columnName);
		}
	return true;
	}

















