/***************************************************************************
    copyright            : (C) 1999 by Frederik Bilhaut
    email                : bilhaut_f@mail.cpod.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "taggedtext.h"

TaggedText::TaggedText(void)
	: root("root"), separator(':'), error(0), message("")
	{
	}
	
char TaggedText::pathSeparator(void) const
	{
	return separator;
	}
	
void TaggedText::setPathSeparator(char sep)
	{
	separator = sep;
	}
	
QString TaggedText::get(const QString &path) const
	{
	return get(path, &root);
	}
	
QString TaggedText::get(const QString &path, const Node *node) const
	{
	if(path.isEmpty())
		return node->value();
	Node *child = node->child(firstTagOfPath(path));
	if(child == 0)
		return "";
	else
		return get(restOfPath(path), child);
	}

		
bool TaggedText::set(const QString &path, const QString &value)
	{
	return set(path, value, &root);
	}
	
QString TaggedText::operator[](const QString &path) const // For reading
	{
	return get(path);
	}
	
QString &TaggedText::operator[](const QString &path) // For writing
	{
	return getValueRef(path, &root);
	}
	
QString &TaggedText::getValueRef(const QString &path, Node *node)
	{
	if(path.isEmpty())
		return node->valueRef();
	QString tag = firstTagOfPath(path);
	Node *child = node->child(tag);
	if(child == 0)
		{
		child = new Node(tag);
		node->addChild(child);
		}
	QString rest = restOfPath(path);
	return getValueRef(rest, child);
	}
	
bool TaggedText::set(const QString &path, const QString &value, Node *node)
	{
	if(path.isEmpty())
		{
		node->setValue(value);
		return true;
		}
	QString tag = firstTagOfPath(path);
	Node *child = node->child(tag);
	if(child == 0)
		{
		child = new Node(tag);
		if(!child) return false;
		node->addChild(child);
		}
	QString rest = restOfPath(path);
	return set(rest, value, child);		
	}
	
QString TaggedText::firstTagOfPath(const QString &path)  const
	{
	int i = path.find(separator);
	if(i == -1)
		return path;
	else
		return path.left(i);
	}
	
QString TaggedText::restOfPath(const QString &path)  const
	{
	int i = path.find(separator);
	if(i == -1)
		return "";
	else
		return path.right(path.length()-i-1);
	}
	
QString TaggedText::addToPath(const QString &path, const QString &elem)  const
	{
	QString res = path;
	if(path == "")
		res = elem;
	else
		res += ":" + elem;
	return res;
	}
	
QString TaggedText::lastFromPath(const QString &path)  const
	{
	int index = path.findRev(separator);
	if(index == -1)
		return path;
	else
		return path.right(path.length()-index-1);
	}

QString TaggedText::removeLastFromPath(const QString &path)  const
	{
	int index = path.findRev(separator);
	if(index == -1)
		return "";
	else
		return path.left(index);
	}
	
QString TaggedText::toString(void)  const
	{
	return root.toString(false);
	}

bool TaggedText::fromString(const QString &source)
	{
	error = 0;
	message = "Ok";
	clear();
	int index = 0;
	int length = source.length();
	QString tag = "";
	QString contents = "";
	QString path = "";
	bool inTag = false;
	const bool Open = true;
	const bool Close = false;
	bool lastTag = Open;
	while(index < length)
		{
		char c = source[index++];
		if(inTag)
			{
			if(c == '>')
				{
				inTag = false;
				if(tag.isEmpty())
					{
					error = 1;
					message.sprintf("Error %d: tag is empty (at position %d).", error, index);
					clear();
					return false;
					}
				else if(tag[0] == '/')
					{
					tag = tag.right(tag.length()-1);
					QString openedTag = lastFromPath(path);
					if(openedTag != tag)
						{
						error = 2;
						message.sprintf("Error %d: last opened tag is '%s' while closed tag is '%s' (at position %d).", error, (const char *)openedTag, (const char *)tag, index);
						clear();
						return false;
						}
					if(lastTag == Open)
						{
						//warning("TaggedTag::fromString(): Adding '%s' at path '%s'.", (const char *)contents, (const char *)path);
						set(path, contents);
						}
					path = removeLastFromPath(path);
					//warning("TaggedTag::fromString(): Closing tag '%s' (current path='%s').", (const char *)tag, (const char *)path);					
					lastTag = Close;
					}
				else
					{
					lastTag = Open;
					path = addToPath(path, tag);
					contents = "";
					//warning("TaggedTag::fromString(): Entering tag '%s' (current path='%s').", (const char *)tag, (const char *)path);					
					}					
				}
			else if(c == '<')
				{
				error = 3;
				message.sprintf("Error %d: '<' found within a tag (at position %d).", error, index);
				clear();
				return false;
				}
			else
				{
				tag += c;
				}
			}
		else
			{
			if(c == '\\')
				{
				c = source[index++];
				contents += c;
				}
			else if(c == '>')
				{
				error = 4;
				message.sprintf("Error %d: not quoted '>' found outside a tag (at position %d).", error, index);
				clear();
				return false;
				}
			else if(c == '<')
				{
				tag = "";
				inTag = true;
				}
			else
				{
				contents += c;
				}
			}
		}
	if(inTag)
		{
		error = 5;
		message.sprintf("Error %d: unexpected end of input while reading a tag (at position %d).", error, index);
		clear();
		return false;
		}
	error = 0;
	message = "Ok";
	return true;
	}
		
void TaggedText::clear(void)
	{
	root.clear();
	}

int TaggedText::errorCode(void)  const
	{
	return error;
	}
	
QString TaggedText::errorMessage(void) const
	{
	return message;
	}
	
// ************ TaggedText::Node

TaggedText::Node::Node(const QString &aname)
	: _name(aname)
	{
	_children.setAutoDelete(true);
	}
	
QString TaggedText::Node::name(void) const
	{
	return _name;
	}
	
QString TaggedText::Node::value(void) const
	{
	return _value;
	}
	
QString &TaggedText::Node::valueRef(void)
	{
	return _value;
	}
	
QString TaggedText::Node::quotedValue(void) const
	{
	QString res = "";
	for(unsigned int i=0; i<_value.length(); i++)
		{
		char c = _value[i];
		if(c == '<' || c == '>' || c == '\\')
			res += '\\';
		res += c;
		}
	return res;
	}
	
TaggedText::Node *TaggedText::Node::child(const QString &childName) const
	{
	for(QListIterator<Node> it(_children); it.current(); ++it)
		if(it.current()->name() == childName)
			return it.current();
	return 0;
	}
	
void TaggedText::Node::setValue(const QString &avalue)
	{
	_value = avalue;
	}
	
void TaggedText::Node::addChild(TaggedText::Node *node)
	{
	_children.append(node);
	}
	
QString TaggedText::Node::toString(bool printTags) const
	{
	QString res = "";
	if(printTags)
		res += "<" + name() + ">";
	if(_children.count() == 0)
		res += quotedValue();
	else
		for(QListIterator<Node> it(_children); it.current(); ++it)
			res += it.current()->toString();
	if(printTags)
		res += "</" + name() + ">";
	return res;
	}
	
void TaggedText::Node::clear(void)
	{
	_value = "";
	_children.clear();
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	

